# -*- coding: utf8 -*-
import sys
import unittest

from .liskov import behave_as, can_substitute, subtype, append_sys_path, constrain, under_constraint

import os.path
__dir__ = os.path.dirname(os.path.abspath(__file__))
append_sys_path(os.path.join(__dir__, 'fixtures'))

import elephant


class LiskovSubstitutionTest(unittest.TestCase):

    def test_can_substitute_decorator_create_new_subtype(self):
        @can_substitute('calc.BasicCalc', 'convert.BasesConvert')
        class ScientificCalc(object):
            pass

        from calc import BasicCalc
        from convert import BasesConvert
        assert issubclass(ScientificCalc, (BasicCalc, BasesConvert))


    def test_behave_as_metaclass_create_new_subtype(self):
        CalcConvert = behave_as('calc.BasicCalc', 'convert.BasesConvert')
        if sys.version < '3':
            class ScientificCalc(object):
                __metaclass__ = CalcConvert
        else:
            ScientificCalc = CalcConvert('ScientificCalc', (object, ), {})

        from calc import BasicCalc
        from convert import BasesConvert
        assert issubclass(ScientificCalc, (BasicCalc, BasesConvert))


    def test_subtypeof_import_a_class_in_nested_scope(self):
        class ScientificCalc(subtype('calc.BasicCalc')):
            pass

        from calc import BasicCalc
        assert issubclass(ScientificCalc, BasicCalc)

class ConstrainedSubtypeTest(unittest.TestCase):
    def test_can_substitute_except_constraints(self):
        @can_substitute('elephant.ElephantTest')
        @under_constraint('test_it_can_be_grey', 'test_it_can_be_white')
        class RoyalElephantTest(object):
            def new_elephant(self, color):
                return elephant.RoyalElephant(color)

        self.__run(RoyalElephantTest)

    def test_behave_as_except_constraints(self):
        ElephantTest = behave_as('elephant.ElephantTest').except_for('test_it_can_be_grey', 'test_it_can_be_white')
        if sys.version < '3':
            class RoyalElephantTest(object):
                __metaclass__ = ElephantTest

                def new_elephant(self, color):
                    return elephant.RoyalElephant(color)
        else:
            def new_elephant(self, color):
                return elephant.RoyalElephant(color)
            RoyalElephantTest = ElephantTest('RoyalElephantTest', (object, ), {'new_elephant': new_elephant})

        self.__run(RoyalElephantTest)

    def test_subtypeof_except_constraints(self):
        class RoyalElephantTest(subtype('elephant.ElephantTest') & constrain('test_it_can_be_grey', 'test_it_can_be_white')):
            def new_elephant(self, color):
                return elephant.RoyalElephant(color)

        self.__run(RoyalElephantTest)

    def __run(self, testCase):
        loader = unittest.TestLoader()
        suite = loader.loadTestsFromTestCase(testCase)
        suite.debug()

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf8 -*-
from setuptools import setup
import os

pkgName = 'liskov'
setup(
    name=pkgName,
    version='0.2.2',
    url='http://www.python.org/pypi/' + pkgName,
    author='Grégory Salvan',
    author_email='apieum@gmail.com',
    license='LGPL',
    description='Small helpers to make inheritance with unit tests',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    include_package_data=True,
    packages=['liskov'],
    zip_safe=True,
)



# -*- coding: utf8 -*-

__all__=['behave_as', 'can_substitute', 'subtype', 'append_sys_path', 'constrain', 'under_constraint']


def behave_as(*bases):
    substitute_bases = _load_bases(bases)

    class Liskov(type):
        constraints = tuple()
        def __new__(cls, name, bases, attrs):
            new_bases = _remove_bases_behaviours(substitute_bases, cls.constraints)
            return type.__new__(cls, name, new_bases + bases, attrs)

        @classmethod
        def except_for(cls, *behaviours):
            cls.constraints = behaviours
            return cls

    return Liskov

def subtype(base):
    mod_name, sep, cls_name = base.rpartition('.')
    module = __import__(mod_name, globals(), locals(), [cls_name])
    return getattr(module, cls_name)

def can_substitute(*bases):
    substitute_bases = _load_bases(bases)

    def liskov(cls, substitute_bases=substitute_bases):
        if hasattr(cls, 'constrained_subtype'):
            substitute_bases = _remove_bases_behaviours(substitute_bases, cls.constrained_subtype)

        bases = (cls, ) + substitute_bases + cls.__bases__
        attrs = dict(cls.__dict__)
        attrs['substitute_bases'] = substitute_bases
        return type(cls.__name__, bases, attrs)

    return liskov

def under_constraint(*behaviours):
    def constraints(cls):
        setattr(cls, 'constrained_subtype', behaviours)
        return cls
    return constraints

class constrain(object):
    def __init__(self, *behaviours):
        self.constraints = behaviours

    def __rand__(self, cls):
        bases = _remove_bases_behaviours(cls.__bases__, self.constraints)
        attrs = _remove_behaviours(cls.__dict__, self.constraints)
        return type(cls.__name__, bases, attrs)

    __ror__ = __radd__ = __rsub__ = __rand__

def append_sys_path(path):
    import sys, os
    if os.path.isfile(path):
        path = os.path.dirname(path)

    sys.path.append(os.path.realpath(path))

def _remove_bases_behaviours(bases, behaviours):
    new_bases = tuple()
    for base in bases:
        base_attrs = base.__dict__
        attrs = _remove_behaviours(base_attrs, behaviours)
        if len(base_attrs) > len(attrs):
            base = type(base.__name__, base.__bases__, attrs)
        new_bases += (base, )
    return new_bases

def _remove_behaviours(attrs, behaviours):
    new_attrs = dict()
    for key, value in list(attrs.items()):
        if key not in behaviours:
            new_attrs[key] = value
    return new_attrs


def _load_bases(bases):
    substitute_bases = tuple()
    for base in bases:
        cls = subtype(base)
        if cls not in substitute_bases:
            substitute_bases+= (cls, )
    return substitute_bases
